#-*- coding: utf-8 -*-
"""
.. autoclass:: SessionMakerExists
    :members:

.. autoclass:: BaseModel
    :members:
"""
from functools import wraps
from sqlalchemy.orm.properties import RelationshipProperty
from sqlalchemy.orm.collections import InstrumentedList

from sqla_helpers import loading
from sqla_helpers.process import process_params
from sqla_helpers.utils import call_if_callable

class SessionMakerExists(Exception):
    """
    Exception raised when a session maker is already registered in :class:`sqla_helpers.base_model.BaseModel`
    """
    def __unicode__(self):
        return u'A session maker is already registered.'

    def __str__(self):
        return 'A session maker is already registered'


class ClassProperty(property):
    """
    Provide a decorator `property` to classemethod
    """
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


def query_operation(decorated_method=None, operation_name=None):
    """
    query_operation process a search on a query with criterion and operators.
    Then, the operation_name operation is called on the query. If operation_name is not set
    the operation_name is taken from the decorated method

    .. code-block:: python

        @query_operation
        def count(cls, *operators, **criterions):
            # This method will call the cls.search(*operators, **criterions)
            # Then , it will call count operation on the result of previous search
            # I.E. : the generated code will be "return cls.search(*operators, **criterions).count()"

        @query_operation(operation_name='one')
        def get(cls, *operators, **criterions):
            # It will process the same thing above, except the function compute on the query will be 'one'
            # I.E. : return cls.search(*operators, **criterions).one()
 
    .. warning:

        This decorataror defines the decorated_method as a classmethod.

    """

    if  decorated_method and operation_name is None:
        operation_name = decorated_method.__name__

    def _wrapper(decorated_method):

        @classmethod
        @wraps(decorated_method)
        def _decorator(querying_class, *operators, **criterions):

            query = querying_class.search(*operators, **criterions)
            method = query.__getattribute__(operation_name)
            return method()

        return _decorator

    if decorated_method is not None:
        return _wrapper(decorated_method)
    else:
        return _wrapper


class BaseModel(object):
    """
    Base Model Class.
    Provide syntactic sugar for getting object from database.
    """

    sessionmaker = None
    process_params = classmethod(process_params)


    @classmethod
    def register_sessionmaker(cls, sessionmaker, force=False):
        """
        Register the function for making session.
        This registered function mustn't have any parameters.
        For a globale session, just put the session as parameter.
        If a session maker is already registered, an exception is raised to avoid conflict.
        But, if you are sure about what your are doing, you can set `force` parameter to True.
        It's not advice.
        """
        if cls.sessionmaker is None or force:
            cls.sessionmaker = staticmethod(sessionmaker)
        else:
            raise SessionMakerExists()	


    @ClassProperty
    @classmethod
    def session(cls):
        """
        Call :attr:`BaseModel.sessionmaker` and returns a new session.

        Don't forget to call  :attr:`BaseModel.sessionmaker_maker` in application's initialization.
        """
        return call_if_callable(cls.sessionmaker)


    @classmethod
    def search(cls, *operator, **criterion):
        """
        Object search with criterions given in arguments.
        Returns a :class:`sqlachemy.orm.query.Query` object.

        Filters can be chained.
        """
        query = cls.session.query(cls)
        # On maintient une liste des classes déjà jointes
        joined_class = []
        clauses = []

        # On itère sur tous les objets operator qu'on a reçu et on process les
        # objets
        for operator in list(operator):
            clauses.append(operator(cls, joined_class))

        # On process les critéres qu'on nous passe directement
        clauses.extend(cls.process_params(joined_class, **criterion))
        query = query.join(*joined_class)
        return query.filter(*clauses)


    @query_operation(operation_name='one')
    def get(cls, *operators, **criterions):
        """
        Returns an object with criterions given in parameters.
        """

    @query_operation
    def all(cls):
        """
        Returns all objects from the same class contained in database.
        """

    @query_operation(operation_name='all')
    def filter(cls, *operators, **criterions):
        """
        Returns a list of objects from a class matching criterions given in parameters.
        """

    @query_operation
    def count(cls, *operators, **criterions):
        """
        Returns the number of objects matched by criterions

        .. code-block:: python

           >>> Treatment.count(status=u'OK')
           8
        """


    @classmethod
    def load(cls, dictionary, hard=False):
        """
        Returns an object from class with attributes got in dictionary's parameters.

        If all the primary keys are found in the dictionary, the object is
        loaded from database. Otherwise, values are set in the loading object.

        .. code-block:: python

            >>> t = Treatment.get(id=1)
            >>> t.name
            'Great Treatment'
            >>> t = Treatment.load({'id': 1, 'name': 'Awesome Treatment'})
            >>> t.name
            'Awesome Treatment'
            >>> session.commit()
            >>> Treatment.get(id=1).name
            'Awesome Treatment'

        If `hard` parameter is True, an exception is raised if a value isn't found
        in parameter's dictionary.
        """

        # On détermine si on doit charger l'instance depuis la base ou non.
        # La décision est prise si tous les attributs qui constitue la clef
        # primaire de l'objet sont trouvés dans le dico. Si oui, on charge
        # depuis la base, sinon on crée une nouvelle instance
        from_db = False
        loading_key = {}
        for attr in cls.__mapper__.primary_key:
            if not attr.key in dictionary:
                from_db = False
                break

            loading_key[attr.key] = dictionary[attr.key]
            from_db = True

        if from_db:
            instance = cls.get(**loading_key)
        else:
            instance = loading.instancied(cls)

        # Une fois l'instance crée , on itère sur toutes les propriétés de
        # la classe à laquelle elle appartient.
        # Et on cherche dans le dictionnaire si elles apparaissent.
        # En mode `hard`, si un clef n'est pas spécifiée on léve une
        # exception, sinon on ignore l'erreur.
        for attr_key, attr_type in cls.__mapper__._props.iteritems():
            try:
                attr_value = dictionary[attr_key]
            except KeyError:
                # En mode soft, on ne relache pas l'erreur
                # En mode hard, oui.
                if hard:
                    raise
            else:
                instance_attr = getattr(instance, attr_key)
                # Si l'attribut sur lequel nous sommes, est une relation,
                # On délégue son chargement à la classe à laquelle l'attrbut
                # appartient.
                # Sinon on se contente de mettre la valeur trouvée dans le
                # dictionnaire, dans l'attribut de la nouvelle instance.
                if isinstance(attr_type, RelationshipProperty):
                    attr_class = attr_type.mapper.class_
                    if isinstance(instance_attr, InstrumentedList):
                        # Si on est sur une liste, on s'attend à avoir une
                        # liste de dico
                        for obj_to_load in attr_value:
                            instance_attr.append(attr_class.load(obj_to_load))
                    else:
                        instance_attr = attr_class.load(attr_value)

                else:
                    instance_attr = attr_value

                setattr(instance, attr_key, instance_attr)

        return instance




    def dump(self, excludes=[], depth=2):
        """
        Returns object as dictionary with dependencies.

        `Depth` limits the recursion.

        IE : With depth set as 1, objects in relations aren't search.

        `excludes` use to exclude unwanted attributes.

        .. code-block:: python

            >>> t = Treatment.get(id=1)
            >>> print json.dumps(t.dump(), indent=4)
            {
                "status": {
                    "id": 1,
                    "name": "Ok"
                },
                "status_id": 1,
                "id": 1,
                "name": "Great Treatment"
            }

            >>> print json.dumps(t.dump(depth=1), indent=4)
            {
                "status_id": 1,
                "id": 1,
                "name": "Great Treatment"
            }
            >>> print json.dumps(t.dump(excludes=['status_id']), indent=4)
            {
                "status": {
                    "id": 1,
                    "name": "Ok"
                },
                "id": 1,
                "name": "Great Treatment"
            }

        """
        res = {}

        # On itére sur les propriétés de classes pour récupérer seulement
        # les attributs déclarer en base pour ne pas exporter les autres attributs
        # Mais on récupère bien la valeur dans l'instance d'objet.
        for attr, attr_type in self.__mapper__._props.iteritems():

            # Si le champ est à exclure on passe au champ suivant
            if attr in excludes:
                continue

            # Récupération de la valeur effective de l'attribut
            instance_attr = getattr(self, attr)
            if isinstance(attr_type, RelationshipProperty):

                # Si on est à la profondeur on ne fait rien
                if depth - 1 <= 0:
                    continue

                if isinstance(instance_attr, InstrumentedList):
                    res[attr] = [ a.dump(depth=depth-1) for a in instance_attr]
                else:
                    res[attr] = instance_attr.dump(depth=depth-1)

            else:
                res[attr] = instance_attr

        return res



# -*- coding: utf-8 -*-
operators = {
    'not': '__ne__',
    'lt': '__lt__',
    'le': '__le__',
    'gt': '__gt__',
    'ge': '__ge__',
    'in': 'in_',
    'like': 'like',
    'ilike': 'ilike',
}
"""
Dictonnary of usasable operators to a `InstrumentedAttribut`.
Keys are operators in :mod:`sqla_helpers` syntax and values are methods name
called by an InstrumentAttribut object.
"""

def process_params(cls, class_found, **kwargs):
    """
    Returns a `SQLAlchemy` criterions list matching :mod:`sqla_helpers` syntax.

    :param:`cls` is the root class providing attributes

    During processing, found attributes are stored  in :param:`class_found`
    parameters. :param:`class_found` is a set.

    Attribute is updated during process.

    .. rubric:: Example

    If a quering on the attribute `name` from a `Treatment` object, the function
    'll be called:

    .. code-block:: python

        >>> class_found = []
        >>> process_params(Treatment, class_found, name='test')
        [<sqlalchemy.sql.expression.BinaryExpression object at 0x22bd3d0>]
        >>> class_found
        []

    In this example, because of the attribute `name` isn't in the related object, the
    :param:`class_found` isn't modified. In other hand a query on `status` attribute :

    .. code-block:: python

        >>> class_found = []
        >>> process_params(Treatment, class_found, status_name='test')
        [<sqlalchemy.sql.expression.BinaryExpression object at 0x22bd3d0>]
        >>> class_found
        [Status]

    """
    criterion = []
    for k, v in kwargs.iteritems():
        # Si il y a des __ dans le paramètre, on souhaite
        # faire une recherche sur un attribut d'une relation
        params = k.split('__')
        # Le dernier élément peut être un opérateur
        if params[-1] in operators.keys():
            op = params.pop()
            operator = operators[op]
        else:
            operator = '__eq__'
        # On récupère le nom de l'attribut de comparaison
        # qui est systèmatiquement en dernier
        comparator_attr_name = params.pop()
        # On garde la classe à partir de laquelle on récupère l'attribut
        # courant.
        klass = cls
        # La boucle permet de récupérer l'attribut dans la classe la plus
        # loin dans les relations.
        # Exemple = [parameter, task, id_task]
        # comparator_attr_name = "id_task"
        # cls.(classe de l'attribut "parameter").(classe de l'attribut
        # "task").id_task)
        for param in params:
            klass = getattr(klass, param).property.mapper.class_
            if klass not in class_found:
                class_found.append(klass)

        instrumented_attr = getattr(klass, comparator_attr_name)
        comparator = getattr(instrumented_attr, operator)
        criterion.append(comparator(v))

    return criterion



from nose import with_setup
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from sqla_helpers.tests.class_test import Treatment, Status


from sqla_helpers.base_model import BaseModel
from sqla_helpers.tests.class_test import metadata

engine = create_engine('sqlite://')
session = sessionmaker(bind=engine)()

def populate():
	BaseModel.register_sessionmaker(session, force=True)
	metadata.create_all(engine)
	status = [Status(u'ok'), Status(u'ko')]
	session.add_all(status)
	session.commit()
	

def unpopulate():
	session.query(Treatment).delete()
	session.query(Status).delete()
	session.commit()


@with_setup(populate, unpopulate)
def test_load():
	status = Status.load({'name': u'plop'})
	assert status.name == u'plop'
	status = Status.load({'id': 1})
	assert status.id == 1 
	assert status.name == u'ok'



from nose.tools import raises
from sqla_helpers.tests.class_test import  Treatment, Status

from sqla_helpers.process import process_params

def test_simple():
    res = process_params(Treatment, [], id=0)
    assert len(res) == 1
    assert str(Treatment.id == 0) == str(res[0])
    res = process_params(Treatment, [], name=0)
    assert len(res) == 1
    assert str(Treatment.name == 0) == str(res[0])


def test_multiple_key():
    res = process_params(Treatment, [], id=0, name='toto')
    assert len(res) == 2


def test_not_operator():
    res = process_params(Treatment, [], id__not=0)
    assert len(res) == 1
    assert str(Treatment.id != 0) == str(res[0])


def test_lt_operator():
    res = process_params(Treatment, [], id__lt=0)
    assert len(res) == 1
    assert str(Treatment.id < 0) == str(res[0])


def test_le_operator():
    res = process_params(Treatment, [], id__le=0)
    assert len(res) == 1
    assert str(Treatment.id <= 0) == str(res[0])

def test_gt_operator():
    res = process_params(Treatment, [], id__gt=0)
    assert len(res) == 1
    assert str(Treatment.id > 0) == str(res[0])

def test_ge_operator():
    res = process_params(Treatment, [], id__ge=0)
    assert len(res) == 1
    assert str(Treatment.id >= 0) == str(res[0])

def test_in_operator():
    res = process_params(Treatment, [], id__in=[0])
    assert len(res) == 1
    assert str(Treatment.id.in_([0])) == str(res[0])


@raises(AttributeError)
def test_unknow_attr():
    process_params(Treatment, [], test='toto')

def test_simple_relation():
    joined_class = []
    res = process_params(Treatment, joined_class, status__name='test')
    assert len(joined_class)
    assert Status in joined_class
    assert str(Status.name == 'test') == str(res[0])

def test_multiple_relation():
    joined_class = []
    process_params(Treatment, joined_class, status__name='test',
                         status__id=0)
    assert len(joined_class) == 1
    assert Status in joined_class



#-*- coding: utf-8 -*-
"""
Some tools to help development
==============================

.. autofunction:: call_if_callable

"""

def call_if_callable(maybe_callable, callable_args=None, callable_kwargs=None, else_=lambda x: x):	
    """
    If `maybe_callable` is callable then it's called with `*callable_args` and `**callable_kwargs`.
    Else the function `else_` is called with `maybe_callable` as argument. `else_` default is the
    identity function, so when maybe_callable is not callable, the function returns `maybe_callable.`

    .. code-block:: python

        >>> call_if_callable(u'test')
        u'test'
        >>> def add2(x):
        ...     return x + 2
        >>> call_if_callable(add2, [4])
        6
        >>> def add_something(x, something=2)
        ...     return  x + something
        >>> call_if_callable(add_something, [4], {'something': 8})
        12
        >>> def concat_plop(string):
        ...        return u'{0}.plop'.format(string)
        >>> call_if_callable(u'test', else_=concat_plop)
        u'test.plop'
    
    :param maybe_callable: function to call if callable
    :param callable_args: List of argument to call with `maybe_callable`
    :param callable_kwargs: Dictonary of (key, value) to call with `maybe_callable`
    :param else_: Function to call if maybe_callable is not callable. One parameter supported : `maybe_callable`
    """

    callable_args = callable_args or []
    callable_kwargs = callable_kwargs or {}

    if hasattr(maybe_callable, '__call__'):
        res = maybe_callable(*callable_args, **callable_kwargs)
    else:
        res = else_(maybe_callable)

    return res



# -*- coding: utf-8 -*-

"""
Abstract Syntatic Tree
======================

.. autoclass:: ASTNode
    :members:
.. autoclass:: AndASTNode
    :members:
.. autoclass:: OrASTNode
    :members:

Q Object
========
.. autoclass:: Q
    :members:
"""

from sqlalchemy import or_, and_, not_
from sqla_helpers.process import process_params


class ASTNode(object):
    """
    A tree's node represent a logic Sqlalchemy operator.
    Contains 2 child tree of a set of criterion.


    Handled criterions are describes in  :mod:`sqla_helpers`.

        >>> ast = AndASTNode(id=4, status__name='foo')

    Criterions 'll be process by the `process_param` function
    during processing of subtree by `__call__` function.

        >>> ast(MyClass, [])
        <sqlalchemy.sql.expression.BinaryExpression object at 0x1f04090>

    If a node contains criterions, the node is a leaf. (meaning the :attr:`ASTNode.operand`
    attribute is not `None`.)

    If node contains children, a recursive processing of children subtrees is done.

    `ASTNode.operator` method is excecuted during the return of `process_param`
    or during the recursive return of children.

    :class: `ASTNode` is an abstract class which doesn't implement :method: ASTNode.operator`.
    """

    def operator(self, *args, **kwargs):
        raise NotImplementedError()


    def __init__(self, lhs=None, rhs=None, **operand):
        self.lhs = lhs
        self.rhs = rhs
        self.operand = operand


    def __call__(self, klass, class_found):
        """
        Process every node.
        """
        # Si l'on a des paramètres bruts, c'est que l'on est une feuille de
        # l'arbre
        # On retourne alors le process des paramètres
        if self.operand:
            clauses = process_params(klass, class_found, **self.operand)
        else:
            # Sinon, nous avons des enfants et l'on retourne l'opération que l'on
            # représente sur le retour des enfants.
            clauses = []
            if self.lhs:
                clauses.append(self.lhs(klass, class_found))
            if self.rhs:
                clauses.append(self.rhs(klass, class_found))

        return self.operator(*clauses)


class OrASTNode(ASTNode):

    def operator(self, *args):
        """
        Excecut a logical or on `SQLAlchemy` criterions.
        """
        return or_(*args)


class AndASTNode(ASTNode):

    def operator(self, *args):
        """
        Excecut a logical and on `SQLAlchemy` criterions.
        """
        return and_(*args)


class NotASTNode(ASTNode):

    def __init__(self, lhs=None, **operand):
        super(NotASTNode, self).__init__(lhs, None, **operand)


    def operator(self, *args):
        """
        Exécute un NON logique sur les critères SQLAlchemy
        """
        return not_(*args)



class Q(object):
    """
    Class representing logical operators with sqla_helpers syntax.

    Q object is only here for processing the `sqla_helpers` syntax
    and implements opertators such as `or`, `and`, `not` ...

    During operations, the :class:`Q` object maintains an AST.
    Each operation return a :class:`Q` object with an updated AST in
    relation the current object and the operating object.

    The :method: `__call__` from :class:`Q` call the :method:`__call__`
    from AST children. The return is an `SQLAlchemy` opertation usable in
    a `Query` object.
    """

    def __init__(self, astnode=None, **kwargs):
        if astnode:
            self.ast = astnode
        else:
            self.ast = AndASTNode(**kwargs)


    def __call__(self, klass, class_found):
        """
        Processing and interpreting AST of current `Q` object.
        """
        return self.ast(klass, class_found)


    def __or__(self, q):
        return Q(OrASTNode(self.ast, q.ast))


    def __and__(self, q):
        return Q(AndASTNode(self.ast, q.ast))


    def __invert__(self):
        return Q(NotASTNode(self.ast))



from functools import wraps

from mock import Mock
from nose.tools import raises

from sqla_helpers.base_model import BaseModel, SessionMakerExists


def save_sessionmaker(fun):
    """
    Guarantee to restore sessionmaker in BaseModel at the end of function called
    """
    @wraps(fun)
    def wrapper(*args, **kwargs):
        old_sessionmaker = BaseModel.sessionmaker
        try:
            res = fun(*args, **kwargs)
        finally:
            BaseModel.sessionmaker = old_sessionmaker    

    return wrapper


@save_sessionmaker
def test_register_simple_session():
    Session = u'session'
    BaseModel.register_sessionmaker(Session)
    assert BaseModel.session == Session


@save_sessionmaker
def test_register_function_session():
    def build_session():
        return u'built_session'

    BaseModel.register_sessionmaker(build_session)
    assert BaseModel.session == u'built_session'


@save_sessionmaker
@raises(SessionMakerExists)
def test_register_session_maker_already_registered():
    BaseModel.register_sessionmaker(u'test')
    BaseModel.register_sessionmaker(u'plop') 


@save_sessionmaker
def test_force_sessionmaker():
    BaseModel.register_sessionmaker(u'test')
    BaseModel.register_sessionmaker(u'plop', force=True) 



from sqla_helpers.utils import call_if_callable

def test_simple_nocallable():
    assert u'test' == call_if_callable(u'test')


def test_simple_callable():
    def simple():
        return u'test'

    assert u'test' == call_if_callable(simple)


def test_simple_callable_args():
    def add2(x):
        return x + 2

    assert 4 == call_if_callable(add2, [2])


def test_simple_callable_kwargs():
    def add_something(x, something=2):
        return x + something 

    assert 4 == call_if_callable(add_something, [2])
    assert 6 == call_if_callable(add_something, [2], {'something': 4})


def test_callable_multiple_args():
    def add_all(x, y, z):
        return x + y + z

    assert 6 == call_if_callable(add_all, [1, 2, 3])


def test_callable_multiple_kwargs():
    def add_all(x, bonii=2, other=4):
        return x + bonii + other

    kwargs = {'bonii': 4, 'other': 8}
    assert 14 == call_if_callable(add_all, [2], kwargs)


def test_callable_else():
    def concat_plop(x):
        return u'{0}.plop'.format(x)

    assert u'test.plop' == call_if_callable(u'test', else_=concat_plop)



#-*- coding: utf-8 -*-

from sqlalchemy.orm.state import InstanceState

def instancied(cls):
    """
    Return a class without use on `__init__`.


    Need more assurance about object creation
    """
    instance = cls.__new__(cls)
    instance._sa_instance_state = InstanceState(instance,
                                                  instance._sa_class_manager)
    return instance



from nose import with_setup
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from sqla_helpers.tests.class_test import Treatment, Status


from sqla_helpers.base_model import BaseModel
from sqla_helpers.logical import Q
from sqla_helpers.tests.class_test import metadata

engine = create_engine('sqlite://')
session = sessionmaker(bind=engine)()

def populate():
    BaseModel.register_sessionmaker(session, force=True)
    metadata.create_all(engine)
    status = [Status(u'ok'), Status(u'ko')]
    session.add_all(status)

    ok  = status[0]
    for i in xrange(10):
        tr = Treatment(u'test {}'.format(i), ok)
        session.add(tr)

    ko = status[1]
    for i in xrange(8):
        tr = Treatment(u'test_ko {}'.format(i), ko)
        session.add(tr)

    session.commit()
	

def unpopulate():
	session.query(Treatment).delete()
	session.query(Status).delete()
	session.commit()


@with_setup(populate, unpopulate)
def test_get():
  tr = Treatment.get(id=1)
  assert isinstance(tr, Treatment)


@with_setup(populate, unpopulate)
def test_filter():
  assert type(Treatment.filter(id=1)) == list
  assert len(Treatment.filter(id=1)) == 1
  assert len(Treatment.filter(~Q(id=1))) == 17
  assert len(Treatment.filter(status__name=u'ok')) == len(Treatment.filter(~Q(status__name=u'ko')))



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

import sqla_helpers

try:
    long_description = open('doc/sqla_helpers.rst').read(),
except IOError:
    long_description = ''

setup(
    name='sqla_helpers',
    version=sqla_helpers.__version__,
    author='Guillaume Camera',
    author_email='camera.g@gmail.com',
    description="Provide some methods for getting objects with SQLAlchemy",
    long_description=long_description,
    url="https://github.com/moumoutte/sqla_helpers/",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: French",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2.7",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    requires=[
        'nose',
        'coverage',
        'mock',
        'sqlalchemy'
    ]
)



# -*- coding: utf-8 -*-
from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base

from sqla_helpers.base_model import BaseModel

DeclarativeModel = declarative_base(cls=BaseModel)
metadata = DeclarativeModel.metadata

class Treatment(DeclarativeModel):
    __tablename__ = 'treatment'
    id = Column('id', Integer, primary_key=True)
    name = Column('name', String)
    status_id = Column('status_id', ForeignKey("status.id"))
    status = relationship('Status', backref='treatments')

    def __init__(self, name, status):
        self.name = name
        self.status = status


class Status(DeclarativeModel):
    __tablename__ = 'status'
    id = Column('id', Integer, primary_key=True)
    name = Column('name', String)

    def __init__(self, name):
        self.name = name
